open Pcre
  
let name = regexp "^\\s*\\[([ !-~]+)\\]\\s*$"
let item = regexp "^\\s*([!-<>-~]+)\\s*=\\s*((\\s*[!-~]+)+)\\s*$";;

module Ordstr = 
struct
  type t = string
  let compare (x:t) (y:t) = 
    String.compare (String.lowercase x) (String.lowercase y)
end

module Strset = Set.Make(Ordstr)

let setOfList list =
  let rec dosetOfList list set = 
    match list with
	a :: tail -> dosetOfList tail (Strset.add a set)
      | []  -> set
  in
    dosetOfList list Strset.empty

let rec read_name fd = 
  let l = input_line fd in
    try
      let o = extract ~rex:name l in
	o.(1)
    with Not_found -> read_name fd

let rec read_body fd tbl =
  let l = input_line fd in
    try
      let o = extract ~rex:item l in
	Hashtbl.add tbl o.(1) o.(2);
	read_body fd tbl;
    with
	Not_found -> tbl
      | End_of_file -> tbl

let rec read_inifile fd tbl =
  try
    let name = read_name fd in
    let body = read_body fd (Hashtbl.create 15) in
      Hashtbl.add tbl name body;
      read_inifile fd tbl
  with End_of_file -> ()

let write_inifile fd tbl = 
  Hashtbl.iter
    (fun k v ->
       output_string fd "[";output_string fd k;output_string fd "]\n";
       (Hashtbl.iter
	  (fun k v ->
	     output_string fd k;output_string fd " = ";output_string fd v;
	     output_string fd "\n") v);
       output_string fd "\n")
    tbl

exception Invalid_section of string
exception Invalid_element of string

class inifile file =
object (self)
  val file = file
  val data = Hashtbl.create 50

  initializer 
  let inch = open_in file in
    read_inifile inch data;
    close_in inch

  method getval sec elt =
    try Hashtbl.find 
      (try (Hashtbl.find data sec)
       with Not_found -> raise (Invalid_section sec))
      elt
    with Not_found -> raise (Invalid_element elt)

  method getaval sec elt =
    try Hashtbl.find_all
      (try (Hashtbl.find data sec)
       with Not_found -> raise (Invalid_section sec))
      elt
    with Not_found -> raise (Invalid_element elt)

  method setval sec elt v =
    (Hashtbl.add 
       (try Hashtbl.find data sec
	with Not_found ->
	  let h = Hashtbl.create 10 in
	    Hashtbl.add data sec h;h)
       elt v);

  method delval sec elt =
    (Hashtbl.remove
       (try Hashtbl.find data sec
	with Not_found -> raise (Invalid_section sec))
       elt)

  method sects =
    (Hashtbl.fold
       (fun k v keys -> k :: keys)
       data [])

  method iter func sec =
    (Hashtbl.iter func 
       (try Hashtbl.find data sec
	with Not_found -> raise (Invalid_section sec)))

  method attrs sec =    
    (Strset.elements
       (setOfList
	  (Hashtbl.fold
	     (fun k v attrs -> k :: attrs)
	     (try Hashtbl.find data sec
	      with Not_found -> raise (Invalid_section sec))
	     [])))

  method save ?(file = file) () = 
    let outch = open_out file in      
      write_inifile outch data;
      flush outch;
end
